// Written by Nathan Callender : 2002
// PVideoWindow
// Version 1.0

#include "PVideoWindow.h"
#include <stdio.h>


PGrabKeysBox::PGrabKeysBox(BRect r, BWindow* w, int32 c)
  : BBox(r)
{
	command = c;
	win = w;
	SetEventMask(B_KEYBOARD_EVENTS);
	SetBorder(B_NO_BORDER);
}

void
PGrabKeysBox::KeyDown(const char* keys, int32 numbytes)
{
	if (numbytes == 1 && win->IsFront())
	{
		// Post the message to the window
		BMessage msg(command);
		msg.AddInt8("KeyPressed", keys[0]);
		win->PostMessage(&msg, win);
	}
}

PVideoBox::PVideoBox(BRect r, uint32 rM)
  : BBox(r, "", rM, B_WILL_DRAW | B_FRAME_EVENTS | 
  		B_FULL_UPDATE_ON_RESIZE, B_NO_BORDER)
{
	bgBitmap = NULL;
}

void
PVideoBox::Draw(BRect)
{
	if (bgBitmap != NULL)
	{
		DrawBitmap(bgBitmap, Bounds());
	}
}



// ===================================================================
// Contructor and Destructor Functions
// ===================================================================

// This is the stand-alone contructor that allows all menus to work.
PVideoWindow::PVideoWindow(BRect r)
  : BWindow(r, "Post Magic Media Player",
        B_TITLED_WINDOW, B_ASYNCHRONOUS_CONTROLS | B_OUTLINE_RESIZE)
{
	InitializeData();
	CreateInterface();    
	runMode = P_RUNMODE_STANDALONE;
}

// The constructor disables the file features (open/close/quit) and
// opens and closes the files automatically given a specified file
PVideoWindow::PVideoWindow(BRect r, entry_ref ref)
  : BWindow(r, "",
        B_TITLED_WINDOW, B_ASYNCHRONOUS_CONTROLS | B_OUTLINE_RESIZE)
{
	InitializeData();
	CreateInterface();
	runMode = P_RUNMODE_INTEGRATED;
	
	// Disabling features
	
	
	// Open the appropriate file ... 
	
	// If fails, give the appropriate error message and close

}

// Cleans up all the dynamically allocated memory.  
PVideoWindow::~PVideoWindow()
{
	delete specialKeysText;
	delete fileInfoText;
/*	delete msgOpenFile;
	delete msgCloseFile;
	delete msgFileInfo;
	delete msgSpecialKeys;
	delete msgSnap25;
	delete msgSnap50;
	delete msgSnap75;
	delete msgSnap100;
	delete msgSnap150;
	delete msgSnap200;
	delete msgSnap400;
	delete msgConstrainProp;
	delete msgPlayStop;
	delete msgStepForward;
	delete msgStepBackward;
	delete msgSetInPoint;
	delete msgSetOutPoint;
	delete msgSliderPos;
	delete msgQuitRequested;
	delete winMessenger;
*/}

// ===================================================================
// Window Hook Functions
// ===================================================================

bool
PVideoWindow::QuitRequested()
{
	// Close any open file
	CloseFile();
    
    // Quit the app
    if (runMode == P_RUNMODE_STANDALONE)
    {
        if (be_app->LockLooper())
    	{	
    		be_app->Quit();
    		be_app->UnlockLooper();
    	}
    }
    
    // Otherwise, quit the looper
    return true;
}

void
PVideoWindow::MessageReceived(BMessage *msg)
{
	entry_ref entryRef;	
	int8 key;
	
    switch (msg->what)
    {
    
    case P_GRABBER_KEY_PRESSED:	msg->FindInt8("KeyPressed", &key);
    							HandleKeyInput(key);
    							break;
   			    
 	case P_OPEN_FILE:			openPanel->Show();
 								break;
 								
 	case P_CLOSE_FILE:			CloseFile();
 								break;
 								
 	case P_SNAP_25:				SnapToFrameScale(0.25);
 								break;
 								
 	case P_SNAP_50:				SnapToFrameScale(0.5);
 								break;
 								
 	case P_SNAP_75:				SnapToFrameScale(0.75);
 								break;
 								
 	case P_SNAP_100:			SnapToFrameScale(1.0);
 								break;
 								
 	case P_SNAP_150:			SnapToFrameScale(1.5);
 								break;
 								
 	case P_SNAP_200:			SnapToFrameScale(2.0);
 								break;
 								
 	case P_SNAP_400:			SnapToFrameScale(4.0);
 								break;
 								
 	case P_CONSTRAIN_PROP:		if (constrainPropMenuItem->IsMarked())
 									constrainPropMenuItem->SetMarked(false);
 								else 
 									constrainPropMenuItem->SetMarked(true);
 								break;
 													
 	case P_SLIDER_POS:			SeekToSliderPosition();
 								break;
 	
 	case P_PLAY_STOP:			if (!isPlaying)	Play();
 								else Stop();
 								break;
 								
 	case P_STEP_BACKWARD:		Step(-stepSize);
 								break;
 								
 	case P_STEP_FORWARD:		Step(stepSize);
 								break;
 								
 	case P_SET_IN_POINT:		SetCurrentAsInPoint();
								break;
								
 	case P_SET_OUT_POINT:		SetCurrentAsOutPoint();
 								break;
 	
 	case P_SPECIAL_KEYS:		vidWinAlert = new BAlert("Special Keys", specialKeysText->String(), "Okay");
								vidWinAlert->Go();
								vidWinAlert = NULL;		
								break;
								
 	case P_FILE_INFO:			vidWinAlert = new BAlert("File Information", fileInfoText->String(), "Okay");
								vidWinAlert->Go();
								vidWinAlert = NULL;		
								break;
								
    case B_REFS_RECEIVED:	
    	// Gets the entry_ref from the Open File Panel			
		if (msg->FindRef("refs", 0, &entryRef) == B_OK)
		{
			OpenFile(entryRef);
		}
    	break;
				   
    default:	    BWindow::MessageReceived(msg);               
    }               
}

void
PVideoWindow::FrameResized(float x, float y)
{	
	if (constrainPropMenuItem->IsMarked() && bufferBitmap != NULL)
	{
		BRect bfrRect = bufferBitmap->Bounds();
		BRect vidRect = videoBox->Bounds();
		float scale = (vidRect.right * vidRect.bottom) / (bfrRect.right * bfrRect.bottom);
		if (!IsFrameProportional())
		{
			SnapToFrameScale(sqrt(scale));
		}
	}
}


// ===================================================================
// Public File Functions
// ===================================================================


bool
PVideoWindow::OpenFile(entry_ref entryRef)
{
	// Make sure there's no open file...
	CloseFile();
	
	// Create a MediaFile to read tracks from
	mediaFile = new BMediaFile(&entryRef);
	if (mediaFile->InitCheck() != B_OK)
	{
		delete mediaFile;
		mediaFile = NULL;

		vidWinAlert = new BAlert("File Error...", "Cannot read this file format!", "Okay");
		vidWinAlert->Go();
		vidWinAlert = NULL;
		return false;
	}
	
	// Find the first audio and video tracks
	BMediaTrack* tmpTrack;
	audioTrack = NULL;
	videoTrack = NULL;
	media_format temf;
	media_format vemf;
	media_format aemf;
	bool foundVideoTrack = false;
	bool foundAudioTrack = false;
	for (int i = 0; i < mediaFile->CountTracks(); i++)
	{
		tmpTrack = mediaFile->TrackAt(i);
		if (tmpTrack != NULL && tmpTrack->InitCheck() == B_OK)
		{
			tmpTrack->EncodedFormat(&temf);
			if ( (!foundVideoTrack) && temf.IsVideo())
			{
				videoTrack = tmpTrack;
				foundVideoTrack = true;
			}	
			else if ( (!foundAudioTrack) && temf.IsAudio()) 
			{
				audioTrack = tmpTrack;
				foundAudioTrack = true;
			}
			else 
			{
				mediaFile->ReleaseTrack(tmpTrack);
			}
		}
	}
	
	fileInfoText->SetTo("File Information: \n\n");
	
	// Output the file opening status...
	if (videoTrack != NULL) printf("Found video track...\n");
	if (audioTrack != NULL) printf("Found audio track...\n");
	
	// Do the video first	
	if (videoTrack != NULL)
	{
		// Set and Get the Decoded Media Format struct
		media_format vdmf;
		vdmf.type = B_MEDIA_RAW_VIDEO;
		vdmf.u.raw_video.display.format = B_RGBA32;
		if (videoTrack->DecodedFormat(&vdmf) != B_OK)
		{
			vdmf.u.raw_video.display.format = B_RGB32;
			if (videoTrack->DecodedFormat(&vdmf) != B_OK)
			{
				vidWinAlert = new BAlert("Codec Error...", "Cannot find an appropriate video codec!", "Okay");
				vidWinAlert->Go();
				vidWinAlert = NULL;
				mediaFile->ReleaseTrack(videoTrack);
				videoTrack = NULL;
				return false;
			}	
		}
		
		// Create the buffer bitmap
		bufferBitmap = new BBitmap(
			BRect(0, 0, vdmf.Width() - 1, vdmf.Height() - 1),
			vdmf.ColorSpace());
	
		// Set field rate
		fieldRate = vdmf.u.raw_video.field_rate;
		
		*fileInfoText << "Video Track:\n"
					  << "\tWidth: " << vdmf.Width() << "\n"
					  << "\tHeight: " << vdmf.Height() << "\n"
					  << "\tField Rate: " << vdmf.u.raw_video.field_rate << "\n"
					  << "\tFields per Frame: " << vdmf.u.raw_video.interlace << "\n"
					  << "\tPixel Aspect Ratio (W/H): " 
					 		<< vdmf.u.raw_video.pixel_width_aspect << "/"
					 		<< vdmf.u.raw_video.pixel_height_aspect << "\n\n"; 
	}

	// Now for the audio 	
	if (audioTrack != NULL)
	{
		media_format admf;
		admf.type = B_MEDIA_RAW_AUDIO;
		if (audioTrack->DecodedFormat(&admf) != B_OK)
		{
			vidWinAlert = new BAlert("Codec Error...", "Cannot find an appropriate audio codec!", "Okay");
			vidWinAlert->Go();
			vidWinAlert = NULL;
			mediaFile->ReleaseTrack(audioTrack);
			audioTrack = NULL;
			return false;	
		}	
	
		// Create a new BSoundPlayer
		const media_raw_audio_format f = admf.u.raw_audio;
		audioPlayer = new BSoundPlayer(&f, "Audio Player", &AudioBufferHook, NULL, (void*) this);
		if (audioPlayer->InitCheck() != B_OK)
		{
			vidWinAlert = new BAlert("Player Error...", "Cannot initialize audio player!", "Okay");
			vidWinAlert->Go();
			vidWinAlert = NULL;		
			mediaFile->ReleaseTrack(audioTrack);
			audioTrack = NULL;
			return false;	
		}		
		audioPlayer->Start();
		audioPlayer->SetHasData(false); 
		
		// Set the info
		sampleRate = admf.u.raw_audio.frame_rate;
		
		// This can go away after testing
		//sampleBufferSize = admf.u.raw_audio.buffer_size
		//	/ (admf.u.raw_audio.format & admf.u.raw_audio.B_AUDIO_SIZE_MASK);
			// This is the buffer size in samples

		*fileInfoText << "Audio Track:\n"
					  << "\tSample Rate: " << admf.u.raw_audio.frame_rate << "\n"
					  << "\tNumber of Channels: " << admf.u.raw_audio.channel_count << "\n\n";
					 
	}
	
	
	// Set the title bar to the file name
	BPath path(&entryRef);
	SetTitle(path.Path());
	
	// Re set the position to the beginning
	positionSlider->SetPosition(0.0);
	
	// Set the limit labels
	if (videoTrack != NULL)
		positionSlider->SetLabel(
			FramesToString(videoTrack->CountFrames()).String());
	else if (audioTrack != NULL)
		positionSlider->SetLabel(
			FramesToString(audioTrack->CountFrames()).String());
}

void 
PVideoWindow::CloseFile()
{	
	// Make sure there's an open file
	if (mediaFile == NULL) return;

	// First we stop the video (if it's playing)
	Stop();
	
	// We delete the bufferBitmap
	if (bufferBitmap != NULL)
	delete bufferBitmap;
	bufferBitmap = NULL;
	
	// Then we release the tracks
	if (videoTrack != NULL)	mediaFile->ReleaseTrack(videoTrack);
	if (audioTrack != NULL)	mediaFile->ReleaseTrack(audioTrack);
	videoTrack = NULL;
	audioTrack = NULL;
	
	// Destroy the audioPlayer (if there is one)
	if (audioPlayer) delete audioPlayer;
	audioPlayer = NULL;
	
	// Close the media file
	mediaFile->CloseFile();
	delete mediaFile;
	mediaFile = NULL;
	
	// Make any changes to the interface
	SetTitle("Post Magic Media Player");
	videoBox->bgBitmap = NULL;
	videoBox->Invalidate();
	positionSlider->SetLabel("");
	positionSlider->SetPosition(0.0);
	positionStrView->SetText("");
	inPointStrView->SetText("");
	outPointStrView->SetText("");
	fileInfoText->SetTo("No file information available!\n\n");
		
	UpdateIfNeeded();
}


// ===================================================================
// Other Public Functions
// ===================================================================

void
PVideoWindow::Play()
{
	if (isPlaying || (audioPlayer == NULL && videoTrack == NULL)) return;
	
	// This must be set for the video and update threads to continue to play.
	// Once this is set to false, they will exit and die. 
	isPlaying = true;

	// Start the audio playing
	if (audioPlayer != NULL)
	{
		audioPlayer->SetHasData(true);
	}

	// Start the video playing
	if (videoTrack != NULL)
	{
		if (audioPlayer == NULL)
			startPerformanceTime = system_time();
				// This slaves the video "time source" to the system time source
		else
			startPerformanceTime = audioPlayer->PerformanceTime();
				// This slaves the video "time source" to the audio player
		
		currentVideoFrame = 0;
			// This is the current video frame from the start
			// of this playing
					 
		thread_id vT = spawn_thread(PlayVideo, "PlayVideo", 100, (void*) this);
		resume_thread(vT);
	}
	
	// Start the updateSlider Thread
	thread_id usT = spawn_thread(Update, "UpdateSlider", 10, (void*) this);
	resume_thread(usT);
	
	// Update interface controls
	if (LockLooper())
	{
		playStopBtn->SetLabel("Stop");
		UnlockLooper();
	}
	
	// Make sure the video box isn't going to draw the bitmap on any window updates	
	videoBox->bgBitmap = NULL;
}

void 
PVideoWindow::Stop()
{
	if (!isPlaying) return;
	
	// This implicitly kills the video thread when it's convenient
	isPlaying = false;

	// This stops the hook function from being called
	if (audioPlayer != NULL)
	{
		audioPlayer->SetHasData(false);
	}
	
	if (LockLooper())
	{
		playStopBtn->SetLabel("Play");
		UnlockLooper();
	}
	
	// Sets the current frame as the one drawn on videoBox redraws
	videoBox->bgBitmap = bufferBitmap;
}

void
PVideoWindow::Step(int64 numFrames)
{
	// Stop the player
	Stop();
	
	// Calculate the frame to seek to
	int64 seekFrame = 0;
	if (videoTrack != NULL)
		seekFrame = videoTrack->CurrentFrame() + numFrames;
	else if (audioTrack != NULL)
		seekFrame = audioTrack->CurrentFrame() + numFrames;
		 
	// ... and then seek
	if (numFrames < 0) 
		SeekToFrame(seekFrame - 1, false);
	else 
		SeekToFrame(seekFrame - 1, true);
		// The megative 1 is there because CurrentFrame returns
		// the next frame that will be returned ... not the
		// one that is being displayed.
}


void
PVideoWindow::SetCurrentAsInPoint()
{
	if (videoTrack != NULL)
 		SetInPoint(videoTrack->CurrentFrame());
 	else if (audioTrack != NULL)
		SetInPoint(audioTrack->CurrentFrame());
}

void
PVideoWindow::SetCurrentAsOutPoint()
{
	if (videoTrack != NULL)
 		SetOutPoint(videoTrack->CurrentFrame());
 	else if (audioTrack != NULL)
		SetOutPoint(audioTrack->CurrentFrame());
}

void
PVideoWindow::SetInPoint(int64 point)
{
	inPointFrame = point;
	if (inPointFrame > outPointFrame)
		outPointFrame = inPointFrame;
	
	BString label;
	label << "IP: " << FramesToString(inPointFrame);
	inPointStrView->SetText(label.String());
	label.SetTo("");
	label << "OP: " << FramesToString(outPointFrame);
	outPointStrView->SetText(label.String());
}


void
PVideoWindow::SetOutPoint(int64 point)
{
	outPointFrame = point;
	if (outPointFrame < inPointFrame)
		inPointFrame = outPointFrame;
	
	BString label;
	label << "IP: " << FramesToString(inPointFrame);
	inPointStrView->SetText(label.String());
	label.SetTo("");
	label << "OP: " << FramesToString(outPointFrame);
	outPointStrView->SetText(label.String());
}

// This provides fast, imperfect time seeking which seeks
// the tracks close, but maybe not perfectly
void 
PVideoWindow::SeekToTime(int64 mediaTime)
{
	// Make sure playback is stopped
	Stop();
	
	// Seek tracks to the specified time
	if (videoTrack != NULL)
	{
		videoTrack->SeekToTime(&mediaTime);
		DisplayNextFrame();
		videoTrack->SeekToTime(&mediaTime);
	}
	
	if (audioTrack != NULL)
	{
		audioTrack->SeekToTime(&mediaTime);
	}
	
	UpdateInterfaceInfo();
}

// This function provides frame accurate seeking for the video track... the audio track
// uses imperfect seeking
void
PVideoWindow::SeekToFrame(int64 iFrame, bool seekForward)
{
	if (iFrame < 0) iFrame = 0;
	int64 oFrame = iFrame;
	if (videoTrack != NULL)
	{
		// There's a video track, so we're seeking the video the correct frame.
		videoTrack->SeekToFrame(&oFrame);
		if (iFrame != oFrame)
		{
			// The seek was imperfect, so we seek to the nearest key frame and read from there
			oFrame = iFrame;
			videoTrack->SeekToFrame(&oFrame, B_MEDIA_SEEK_CLOSEST_BACKWARD);
			
			// Now, oFrame should be less than iFrame.  If not, then just return;
			if (oFrame > iFrame) return;
			
			for (int i = 0; i < iFrame - oFrame; i++)
			{
				DropNextFrame();
			}
		}
		
		// Now, the next frame is the frame we want to display
		DisplayNextFrame();
	
		// Now we want to seek the audio to the right place, but it doesn't have to be exact
		if (audioTrack != NULL)
		{
			int64 iTime = videoTrack->CurrentTime();
			audioTrack->SeekToTime(&iTime);
		}
	}
	else if (audioTrack != NULL)
	{
		// This is a poor, broken implementation, but the real one will follow
		// the same pattern as the video frame-accurate seek
		if (seekForward)
			audioTrack->SeekToFrame(&iFrame, B_MEDIA_SEEK_CLOSEST_FORWARD); 
		else
			audioTrack->SeekToFrame(&iFrame, B_MEDIA_SEEK_CLOSEST_BACKWARD);
	}

	UpdateInterfaceInfo();
}



void
PVideoWindow::SnapToFrameScale(float frameScale)
{
	float winHeight = 0.0;
	float winWidth = 0.0;
	
	if (bufferBitmap != NULL)
	{
		if (frameScale < GetMinFrameScale())
			frameScale = GetMinFrameScale();
			
		BRect bfrRect = bufferBitmap->Bounds();
		BRect vidRect = videoBox->Bounds();
		
		winWidth = Bounds().right - vidRect.right +
			(bfrRect.right * frameScale);
		
		winHeight = Bounds().bottom - vidRect.bottom +
			(bfrRect.bottom * frameScale);
				
		ResizeTo(winWidth, winHeight);		
	}
}

float
PVideoWindow::GetMinFrameScale()
{
	float minHScale = 0.0;
	float minVScale = 0.0;
	if (bufferBitmap != NULL)
	{
		minHScale = (230 - (Bounds().right - videoBox->Bounds().right))
				 	/ bufferBitmap->Bounds().right;
		minVScale = (150 - (Bounds().bottom - videoBox->Bounds().bottom))
					/ bufferBitmap->Bounds().bottom;

	}
	
	if (minHScale > minVScale) return minHScale;
	return minVScale;
}

float
PVideoWindow::GetCurrentFrameScale()
{
	if (bufferBitmap == NULL) return 0.0;
	
	BRect bfrRect = bufferBitmap->Bounds();
	BRect vidRect = videoBox->Bounds();
	float scale = (vidRect.right * vidRect.bottom) / (bfrRect.right * bfrRect.bottom);
	return sqrt(scale);
}

bool
PVideoWindow::IsFrameProportional()
{
	if (bufferBitmap == NULL) return true;
	
	float ratio = (videoBox->Bounds().right / videoBox->Bounds().bottom) -
			      (bufferBitmap->Bounds().right / bufferBitmap->Bounds().bottom);
	
	if (ratio < 0.0) ratio = -ratio;
	if (ratio < 0.01) return true;
	return false;
}





// ===================================================================
// Private Initialization Functions
// ===================================================================

void
PVideoWindow::InitializeData()
{
	vidWinAlert = NULL;
	openPanel = NULL;
	    
    isPlaying = false;
    mediaFile = NULL;
    videoTrack = NULL;
    audioTrack = NULL;
    
    fieldRate = 0.0;
 	sampleRate = 0.0;
    bufferBitmap = NULL;
    currentVideoFrame = 0;
    startPerformanceTime = 0;
        
    audioPlayer = NULL;
    
    inPointFrame = 0;
    outPointFrame = 0;
    stepSize = 5;
    settingStepSize = false;
    
	msgOpenFile 		= new BMessage(P_OPEN_FILE);
	msgCloseFile 		= new BMessage(P_CLOSE_FILE);
	msgFileInfo 		= new BMessage(P_FILE_INFO);
	msgSpecialKeys 		= new BMessage(P_SPECIAL_KEYS);
	msgSnap25 			= new BMessage(P_SNAP_25);
	msgSnap50 			= new BMessage(P_SNAP_50);
	msgSnap75 			= new BMessage(P_SNAP_75);
	msgSnap100 			= new BMessage(P_SNAP_100);
	msgSnap150 			= new BMessage(P_SNAP_150);
	msgSnap200 			= new BMessage(P_SNAP_200);
	msgSnap400 			= new BMessage(P_SNAP_400);
	msgConstrainProp 	= new BMessage(P_CONSTRAIN_PROP);
	msgPlayStop 		= new BMessage(P_PLAY_STOP);
	msgStepForward 		= new BMessage(P_STEP_FORWARD);
	msgStepBackward 	= new BMessage(P_STEP_BACKWARD);
	msgSetInPoint 		= new BMessage(P_SET_IN_POINT);
	msgSetOutPoint 		= new BMessage(P_SET_OUT_POINT);
	msgSliderPos 		= new BMessage(P_SLIDER_POS);
	msgQuitRequested	= new BMessage(B_QUIT_REQUESTED);
	winMessenger 		= new BMessenger(NULL, this, NULL);
}

void
PVideoWindow::CreateInterface()
{
	// Set up the key grabber
	PGrabKeysBox* keyGrabber = new PGrabKeysBox(
		BRect(-20, -20, -18, -18), this, P_GRABBER_KEY_PRESSED);
	AddChild(keyGrabber);

	// Create Menus and Menu bars
    BMenu* vidFileMenu = new BMenu("File");
    vidFileMenu->AddItem(new BMenuItem("Open File", msgOpenFile, 'O'));
    vidFileMenu->AddItem(new BMenuItem("Close File", msgCloseFile, 'C'));
	vidFileMenu->AddSeparatorItem();
	vidFileMenu->AddItem(new BMenuItem("Show File Info", msgFileInfo, 'I'));
	vidFileMenu->AddItem(new BMenuItem("Show Special Keys", msgSpecialKeys, 'K'));
    vidFileMenu->AddSeparatorItem();
    vidFileMenu->AddItem(new BMenuItem("Quit", msgQuitRequested, 'Q'));
    
    BMenu* vidFrameMenu = new BMenu("Frame");
    vidFrameMenu->AddItem(new BMenuItem("Snap to 25%", msgSnap25, '1'));
    vidFrameMenu->AddItem(new BMenuItem("Snap to 50%", msgSnap50, '2'));
    vidFrameMenu->AddItem(new BMenuItem("Snap to 75%", msgSnap75, '3'));
    vidFrameMenu->AddItem(new BMenuItem("Snap to 100%", msgSnap100, '4'));
    vidFrameMenu->AddItem(new BMenuItem("Snap to 150%", msgSnap150, '5'));
    vidFrameMenu->AddItem(new BMenuItem("Snap to 200%", msgSnap200, '6'));
    vidFrameMenu->AddItem(new BMenuItem("Snap to 400%", msgSnap400, '7'));
    vidFrameMenu->AddSeparatorItem();
    constrainPropMenuItem = new BMenuItem("Constrain Proportions", msgConstrainProp);
    vidFrameMenu->AddItem(constrainPropMenuItem);
    constrainPropMenuItem->SetMarked(true);
    
    BMenuBar* vidWinMenuBar = new BMenuBar(BRect(0, 0, 100, 100), "");
    vidWinMenuBar->AddItem(vidFileMenu);
    vidWinMenuBar->AddItem(vidFrameMenu);
    AddChild(vidWinMenuBar); 

	// Create the info Box
	BRect r = Bounds();
	r.top = vidWinMenuBar->Frame().bottom + 1;
	r.bottom = r.top + 16;
	infoBox = new BBox(r, "", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 
		B_WILL_DRAW | B_FRAME_EVENTS, B_PLAIN_BORDER);

	// Create the info string views
	r = infoBox->Bounds();
	r.top += 1;
	r.bottom -= 1;
	r.left += 5;
	r.right = r.left + 75;
	positionStrView = new BStringView(r, "", "");

	r = infoBox->Bounds();
	r.top += 1;
	r.bottom -= 1;
	r.right -= 5;
	r.left  = r.right - 85;
	outPointStrView = new BStringView(r, "", "", B_FOLLOW_RIGHT | B_FOLLOW_TOP);
	outPointStrView->SetAlignment(B_ALIGN_RIGHT);
	
	r.right = r.left - 5;
	r.left = r.right - 85;
	inPointStrView = new BStringView(r, "", "", B_FOLLOW_RIGHT | B_FOLLOW_TOP);
	inPointStrView->SetAlignment(B_ALIGN_RIGHT);
	
	infoBox->AddChild(positionStrView);
	infoBox->AddChild(inPointStrView);
	infoBox->AddChild(outPointStrView);
	
	// Create the video box
	r = Bounds();
	r.top = infoBox->Frame().bottom + 1;
	r.bottom -= 70; 
	videoBox = new PVideoBox(r, B_FOLLOW_ALL_SIDES);

	// Create the control box    
    r = Bounds();
    r.top = videoBox->Frame().bottom + 1;    
    controlBox = new BBox(r, "", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM,
    	B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE_JUMP, B_PLAIN_BORDER);


	// Creat the position slider
	r = controlBox->Bounds();
	r.InsetBy(5, 5);
	r.bottom = r.top + 30;	
    positionSlider = new BSlider(r, "", "", msgSliderPos, 0, 1000,
    	B_TRIANGLE_THUMB, B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM, 
    	B_WILL_DRAW | B_FRAME_EVENTS);
    positionSlider->SetHashMarks(B_HASH_MARKS_TOP);
    positionSlider->SetHashMarkCount(25);
    controlBox->AddChild(positionSlider);   
    positionSlider->SetModificationMessage(new BMessage(P_SLIDER_POS));
    positionSlider->SetSnoozeAmount(1000000LL / 5);

	// Creat the control buttons
	r = controlBox->Bounds();
	r.top = positionSlider->Frame().bottom + 5;
	r.bottom = r.top + 20;
	r.left += 5;
	r.right = r.left + 40;
	stepBackBtn = new BButton(r, "<", "<", msgStepBackward, B_FOLLOW_BOTTOM);
	
	r.left = r.right + 5;
	r.right = r.left + 40;
	playStopBtn = new BButton(r, "Play", "Play", msgPlayStop, B_FOLLOW_BOTTOM);
	
	r.left = r.right + 5;
	r.right = r.left + 40;
	stepForwardBtn = new BButton(r, ">", ">", msgStepForward, B_FOLLOW_BOTTOM);
	

	r.right = controlBox->Bounds().right - 5;
	r.left = r.right - 40;
	setOutPointBtn = new BButton(r, "}", "}", msgSetOutPoint, B_FOLLOW_BOTTOM | B_FOLLOW_RIGHT);

	r.right = r.left - 5;
	r.left = r.right - 40;
	setInPointBtn = new BButton(r, "{", "{", msgSetInPoint, B_FOLLOW_BOTTOM | B_FOLLOW_RIGHT);

	controlBox->AddChild(stepBackBtn);
	controlBox->AddChild(playStopBtn);
	controlBox->AddChild(stepForwardBtn);
	controlBox->AddChild(setInPointBtn);
	controlBox->AddChild(setOutPointBtn);
	
    AddChild(controlBox);
	AddChild(videoBox);   
	AddChild(infoBox);
		
	// Create the Open File Panel
    openPanel = new BFilePanel(B_OPEN_PANEL, winMessenger, NULL, B_FILE_NODE, false);
 
 	// Configure the window
 	SetLook(B_FLOATING_WINDOW_LOOK);  
 	SetSizeLimits(250, 6000, 150, 6000);
 	
 	fileInfoText = new BString("No file information available!\n\n");
 	specialKeysText = new BString("Special Keys:\n\n");
 	*specialKeysText
 		<< "Space \tPlay Media File\n\n" 
 		<< "+ or = \tEnlarge Window\n"
 		<< "- or _ \tShrink Window\n\n"
 		<< "< or , \tSeek Backward\n"
 		<< "> or . \tSeek Forward\n\n"
 		<< "{ or [     \tSet In Point\n"
 		<< "} or ]     \tSet Out Point\n\n"
 		<< "S folowed by a number sets the step size.";
}


// ===================================================================
// Private Play Functions
// ===================================================================


int32
PVideoWindow::Update(void * data)
{
	PVideoWindow* thisWin = (PVideoWindow*) data;

	while (thisWin->isPlaying)
	{
		thisWin->UpdateInterfaceInfo();

		if (thisWin->videoTrack != NULL)
		{
			if (thisWin->videoTrack->CurrentFrame() >=
			    thisWin->videoTrack->CountFrames())
			    thisWin->Stop();
		}
		else if (thisWin->audioTrack != NULL)
		{
			if (thisWin->audioTrack->CurrentTime() >=
			    thisWin->audioTrack->Duration())
			    thisWin->Stop(); 
		}
		snooze(1000000LL / 5);	
	}
	return 0;
}

int32
PVideoWindow::PlayVideo(void * data)
{
	PVideoWindow* thisWin = (PVideoWindow*) data;
	bool displayNextFrame = true;
	int64 snoozeTime = 0;
	
	while (thisWin->isPlaying)
	{
		if (displayNextFrame)
			thisWin->DisplayNextFrame();
		else
		{
			thisWin->DropNextFrame();
			displayNextFrame = true;
		}
		
		thisWin->currentVideoFrame++;

		if (thisWin->audioPlayer != NULL)
		{
			snoozeTime = (int64) thisWin->startPerformanceTime + 
		    	(thisWin->currentVideoFrame * 1000000LL / thisWin->fieldRate) - 
		    	thisWin->audioPlayer->PerformanceTime();
			snooze( snoozeTime );
		}
		else
		{
			snoozeTime = (int64) thisWin->startPerformanceTime + 
		    	(thisWin->currentVideoFrame * 1000000LL / thisWin->fieldRate) - 
		    	system_time();
			snooze ( snoozeTime );		
		}

		if (snoozeTime < 0)	displayNextFrame = false;
	}

	return 0;	
}

void
PVideoWindow::AudioBufferHook(void* data, void* audioBfr, size_t, const media_raw_audio_format &)
{
	PVideoWindow* thisWin = (PVideoWindow*) data;
	if (thisWin->audioTrack == NULL) return;
	int64 outFrameCount = 0;
	thisWin->audioTrack->ReadFrames(audioBfr, &outFrameCount);
}

void
PVideoWindow::DisplayNextFrame()
{
	if (videoTrack == NULL) return;
	
	if (videoBox->Window()->LockLooper())
	{	
		int64 outFrameCount = 0;
		videoTrack->ReadFrames(bufferBitmap->Bits(), &outFrameCount);
		
		videoBox->DrawBitmap(bufferBitmap, videoBox->Bounds());
		videoBox->Window()->UnlockLooper();
	}
}

void
PVideoWindow::DropNextFrame()
{
	if (videoTrack == NULL) return;
	
	int64 outFrameCount = 0;
	media_decode_info mdi;
	mdi.time_to_decode = 0;
	videoTrack->ReadFrames(bufferBitmap->Bits(), &outFrameCount, NULL, &mdi);
}





// ===================================================================
// Other Private Functions
// ===================================================================

// This function updates the interface from the video or audio track locations
void
PVideoWindow::UpdateInterfaceInfo()
{
	float sliderPos = 0.0;
	int64 currentFrame = 0;
	
	if (videoTrack != NULL)
	{
		// We'll use the video track as the master track
		sliderPos = (float) videoTrack->CurrentFrame()
				/ (float) videoTrack->CountFrames();
	
		currentFrame = videoTrack->CurrentFrame();	
	}
	else if (audioTrack != NULL)
	{
		sliderPos = (float) audioTrack->CurrentFrame()
				/ (float) audioTrack->CountFrames();
	
		currentFrame = audioTrack->CurrentFrame();	
	}
				
	if (LockLooper())
	{
		positionSlider->SetPosition(sliderPos);
		positionStrView->SetText(FramesToString(currentFrame).String());
		UnlockLooper();
	}
}

void
PVideoWindow::SeekToSliderPosition()
{
	// Find the media time that the slider currently represents
	int64 sliderMediaTime = 0;
	if (videoTrack != NULL)
	{
		sliderMediaTime = videoTrack->Duration() * positionSlider->Position();	
	} 
	else if (audioTrack != NULL)
	{
		sliderMediaTime = audioTrack->Duration() * positionSlider->Position();	
	}
	else return;
	
	// Seek all tracks to it
	SeekToTime(sliderMediaTime);
}

BString
PVideoWindow::FramesToString(int64 frames)
{
	if (videoTrack == NULL && audioTrack == NULL) return BString("");
	
	// Make sure to check if we should report in audio time or video time
	bool onlyAudio = (videoTrack == NULL);	

	BString returnString("");
	BString hoursString("");
	BString minutesString("");
	BString secondsString("");
	BString framesString("");
	int64 seconds = 0;
	int64 hours = 0;
	int64 minutes = 0;
	float fRate = 1.0;
	
	if (videoTrack != NULL)
		fRate = fieldRate;
	else if (audioTrack != NULL)
		fRate = sampleRate;
					
	hours = frames / fRate / 60 / 60;
	minutes = (frames / fRate / 60) - (hours * 60);
	seconds = (frames / fRate) - (hours * 60 * 60) - (minutes * 60);
	frames = frames - (((hours * 60 * 60) + (minutes * 60) + seconds) * fRate);

	if (onlyAudio) frames = frames * 1000 / fRate;			
	hoursString << hours;
	minutesString << minutes;
	secondsString << seconds;
	framesString << frames;
			
	if (hoursString.CountChars() < 2) hoursString.Prepend("0"); 
	if (minutesString.CountChars() < 2) minutesString.Prepend("0"); 
	if (secondsString.CountChars() < 2) secondsString.Prepend("0"); 
	if (framesString.CountChars() < 2) framesString.Prepend("0"); 

	if (onlyAudio && framesString.CountChars() < 3) framesString.Prepend("0");
		// Audio needs a third place
					
	returnString << hoursString << ":" << minutesString << ":"
			     << secondsString;
			     
	if (onlyAudio) returnString << "." << framesString;
	else returnString << ":" << framesString;

	return returnString;
}

void
PVideoWindow::HandleKeyInput(char key)
{
	float currentScale = GetCurrentFrameScale();
	
	if (settingStepSize) switch (key)
	{
		case '0' :		stepSize = (stepSize * 10) + 0;
						break;
						
		case '1' :		stepSize = (stepSize * 10) + 1;
						break;
						
		case '2' :		stepSize = (stepSize * 10) + 2;
						break;
						
		case '3' :		stepSize = (stepSize * 10) + 3;
						break;
						
		case '4' :		stepSize = (stepSize * 10) + 4;
						break;
						
		case '5' :		stepSize = (stepSize * 10) + 5;
						break;
						
		case '6' :		stepSize = (stepSize * 10) + 6;
						break;
						
		case '7' :		stepSize = (stepSize * 10) + 7;
						break;
						
		case '8' :		stepSize = (stepSize * 10) + 8;
						break;
		
		case '9' :		stepSize = (stepSize * 10) + 9;
						break;
	
		default  :		settingStepSize = false;
						if (stepSize == 0)
							stepSize = oldStepSize;
							
						if (videoTrack != NULL)
							positionSlider->SetLabel(
								FramesToString(videoTrack->CountFrames()).String());
						else if (audioTrack != NULL)
							positionSlider->SetLabel(
								FramesToString(audioTrack->CountFrames()).String());
						else 
							positionSlider->SetLabel("");
							
						break;
	}
	
	if (settingStepSize)
	{
		BString str("Step size in buffer frames: ");
		str << stepSize;
		positionSlider->SetLabel(str.String());
	}
	
	switch (key)
	{
		case '+' :
		case '=' :		currentScale = currentScale * 1.25;
						if (currentScale < 0.1) currentScale = 0.1;
						if (currentScale > 10.0) currentScale = 10.0;
						SnapToFrameScale(currentScale);
						break;
						
		case '-' :
		case '_' :		currentScale = currentScale / 1.25;
						if (currentScale < 0.1) currentScale = 0.1;
						if (currentScale > 10.0) currentScale = 10.0;
						SnapToFrameScale(currentScale);
						break;
						
		case '<' :
		case ',' :		Step(-stepSize);
						break;
						
		case '>' :
		case '.' :		Step(stepSize);
						break;
						
		case '{' :
		case '[' :		SetCurrentAsInPoint();
						break;

		case '}' :
		case ']' :		SetCurrentAsOutPoint();
						break;
						
		case ' ' :		if (!isPlaying)	Play();
 							else Stop();
 						break;
 						
 		case 'S' :
 		case 's' :		settingStepSize = true;
 						oldStepSize = stepSize;
 						stepSize = 0;
 						positionSlider->SetLabel("Step size in buffer frames:");
 						break;
						
		default:		break;
	}			
}














