//  Written by Nathan Callender : 2002
//  PVideoWindow
//  Version 1.0


#include <AppKit.h>
#include <InterfaceKit.h>
#include <TranslationKit.h>
#include <StorageKit.h>
#include <MediaKit.h>
#include <KernelKit.h>

#ifndef _PVIDEOWINDOW_
#define _PVIDEOWINDOW_

class PGrabKeysBox : public BBox
{
	public:
								PGrabKeysBox(BRect, BWindow*, int32);
		virtual void			KeyDown(const char*, int32);
		BWindow*				win;
		int32					command;
};

class PVideoBox: public BBox
{
	public:					
								PVideoBox(BRect, uint32 resizingMode);
	
		virtual void			Draw(BRect);
		BBitmap*				bgBitmap;
};

class PVideoWindow : public BWindow
{
    public:
        						PVideoWindow(BRect);
        							// This is for stand-alone mode where all
        							// functions are turned on
        							
        						PVideoWindow(BRect, entry_ref);
        							// The following are turned off in this mode
        							// - Open File
        							// - Close File
        							// - Quit (it just closest the window)
        							
        						~PVideoWindow();

        virtual bool 			QuitRequested();
        virtual void    		MessageReceived(BMessage*);
        virtual void			FrameResized(float, float);

		bool					OpenFile(entry_ref);
		void					CloseFile();

		void					Play();
		void					Stop();
		void					Step(int64 numFrames);
		void					SetCurrentAsInPoint();
		void					SetCurrentAsOutPoint();        
        
        void					SetInPoint(int64 mediaFrame);
        void					SetOutPoint(int64 mediaFrame);

		void					SeekToTime(int64 mediaTime);
		void					SeekToFrame(int64 mediaFrame, bool seekForward);
		
		void					SnapToFrameScale(float);	
		float					GetMinFrameScale();
		float					GetCurrentFrameScale();
		bool					IsFrameProportional();


		enum {
			P_OPEN_FILE,
			P_CLOSE_FILE,
			P_FILE_INFO,
			P_SPECIAL_KEYS,
			P_SNAP_25,
			P_SNAP_50,
			P_SNAP_75,
			P_SNAP_100,
			P_SNAP_150,
			P_SNAP_200,
			P_SNAP_400,
			P_CONSTRAIN_PROP,
			P_PLAY_STOP,
			P_STEP_FORWARD,
			P_STEP_BACKWARD,
			P_SET_IN_POINT,
			P_SET_OUT_POINT,
			P_SLIDER_POS,
			P_GRABBER_KEY_PRESSED,
			P_RUNMODE_STANDALONE,
			P_RUNMODE_INTEGRATED
		};
		
	private:

		// The window's messages
		BMessage*				msgOpenFile;
		BMessage*				msgCloseFile;
		BMessage*				msgFileInfo;
		BMessage*				msgSpecialKeys;
		BMessage*				msgSnap25;
		BMessage*				msgSnap50;
		BMessage*				msgSnap75;
		BMessage*				msgSnap100;
		BMessage*				msgSnap150;
		BMessage*				msgSnap200;
		BMessage*				msgSnap400;
		BMessage*				msgConstrainProp;
		BMessage*				msgPlayStop;
		BMessage*				msgStepForward;
		BMessage*				msgStepBackward;
		BMessage*				msgSetInPoint;
		BMessage*				msgSetOutPoint;
		BMessage*				msgSliderPos;
		BMessage*				msgQuitRequested;
		BMessenger*				winMessenger;

		// Initialization Functions
		void					InitializeData();
		void					CreateInterface();
		
		// Play Functions
		static int32			Update(void *);
		static int32			PlayVideo(void *);
		static void				AudioBufferHook(void*, void*, size_t,
									const media_raw_audio_format &);
		void					DisplayNextFrame();
		void					DropNextFrame();
        
        // Generic Functions
		void					UpdateInterfaceInfo();
		void					SeekToSliderPosition();
        BString					FramesToString(int64);
        void					HandleKeyInput(char);
        		
		// Interface controls
		BButton*				stepBackBtn;
        BButton*				stepForwardBtn;
        BButton*				playStopBtn;
		BButton*				setInPointBtn;
		BButton*				setOutPointBtn;
        BSlider*				positionSlider;
       	BStringView*			positionStrView;
       	BStringView*			inPointStrView;
       	BStringView*			outPointStrView;
       	BMenuItem*				constrainPropMenuItem;
       	PVideoBox*				videoBox;
       	BBox*					controlBox;
       	BBox*					infoBox;
        
        // Panels and Alerts
        BAlert*					vidWinAlert;
        BFilePanel*				openPanel;
        BString*				specialKeysText;
        BString*				fileInfoText;
        
        // Play objects
        bool					isPlaying;
        BMediaFile*				mediaFile;
        BMediaTrack*			videoTrack;
        BMediaTrack*			audioTrack;

		// Video objects
        float					fieldRate;
        BBitmap*				bufferBitmap;
        int64					currentVideoFrame;
        int64					startPerformanceTime;
        
        // Audio objects
        BSoundPlayer*			audioPlayer;
        float					sampleRate;
        
        //Generic Objects
        int64					inPointFrame;
        int64					outPointFrame;
        int32					runMode;        
        int32					stepSize;
        int32					oldStepSize;
        bool					settingStepSize;
};


#endif